# mypy: disable-error-code="union-attr"
import msgspec
from google.oauth2._service_account_async import Credentials

from lueur.make_id import make_id
from lueur.models import GCPMeta, Resource
from lueur.platform.gcp.client import AuthorizedSession, Client

__all__ = ["explore_addresses"]


async def explore_addresses(
    project: str, location: str | None = None, creds: Credentials | None = None
) -> list[Resource]:
    resources = []

    async with Client("https://compute.googleapis.com", creds) as c:
        if location:
            addresses = await explore_regional_addresses(c, project, location)
            resources.extend(addresses)
        else:
            addresses = await explore_global_addresses(c, project)
            resources.extend(addresses)

    return resources


###############################################################################
# Private functions
###############################################################################
async def explore_global_addresses(
    c: AuthorizedSession,
    project: str,
) -> list[Resource]:
    response = await c.get(f"/compute/v1/projects/{project}/global/addresses")

    addresses = msgspec.json.decode(response.content)

    results = []
    for address in addresses.get("items", []):
        self_link = address["selfLink"]
        name = address["name"]
        display = name

        results.append(
            Resource(
                id=make_id(address["id"]),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="global-addresses",
                    project=project,
                    self_link=self_link,
                    platform="gcp",
                    category="loadbalancer",
                ),
                struct=address,
            )
        )

    return results


async def explore_regional_addresses(
    c: AuthorizedSession, project: str, location: str
) -> list[Resource]:
    response = await c.get(
        f"/compute/v1/projects/{project}/regions/{location}/addresses"
    )

    addresses = msgspec.json.decode(response.content)

    results = []
    for address in addresses.get("items", []):
        self_link = address["selfLink"]
        name = address["name"]
        display = name

        results.append(
            Resource(
                id=make_id(address["id"]),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="regional-addresses",
                    project=project,
                    region=location,
                    self_link=self_link,
                    platform="gcp",
                    category="loadbalancer",
                ),
                struct=address,
            )
        )

    return results



from concurrent.futures import ThreadPoolExecutor

from lueur.make_id import make_id
from lueur.models import AWSMeta, Resource
from lueur.platform.aws.client import Client

__all__ = ["explore_eks"]


def explore_eks(region: str) -> list[Resource]:
    resources = []

    clusters = explore_clusters(region)
    resources.extend(clusters)

    if not clusters:
        return resources

    futures = []
    with ThreadPoolExecutor(max_workers=len(clusters)) as executor:
        for c in clusters:
            futures.append(
                executor.submit(explore_nodegroups, region, c.meta.name)
            )

    for future in futures:
        resources.extend(future.result())

    return resources


###############################################################################
# Private functions
###############################################################################
def explore_clusters(region: str) -> list[Resource]:
    results = []

    with Client("eks", region) as c:
        clusters = c.list_clusters()

        for name in clusters["clusters"]:
            cluster = c.describe_cluster(name)
            results.append(
                Resource(
                    id=make_id(cluster["arn"]),
                    meta=AWSMeta(
                        name=cluster["name"],
                        display=cluster["name"],
                        kind="cluster",
                        platform="aws",
                        region=region,
                        category="compute",
                    ),
                    struct=cluster,
                )
            )

    return results


def explore_nodegroups(region: str, cluster_name: str) -> list[Resource]:
    with Client("eks", region) as c:
        nodegroups = c.list_nodegroups(clusterName=cluster_name)

        results = []

        for ngname in nodegroups["nodegroups"]:
            ng = c.describe_nodegroup(
                clusterName=cluster_name, nodegroupName=ngname
            )

            results.append(
                Resource(
                    id=make_id(ng["nodegroup"]["nodegroupArn"]),
                    meta=AWSMeta(
                        name=ngname,
                        display=ngname,
                        kind="nodegroup",
                        platform="aws",
                        region=region,
                        category="compute",
                    ),
                    struct=ng,
                )
            )

        return results



# mypy: disable-error-code="call-arg"
import msgspec
from kubernetes import client

from lueur.make_id import make_id
from lueur.models import K8SMeta, Resource
from lueur.platform.k8s.client import AsyncClient, Client

__all__ = ["explore_network_policy"]


async def explore_network_policy() -> list[Resource]:
    resources = []

    async with Client(client.NetworkingV1Api) as c:
        policies = await explore_network_policies(c)
        resources.extend(policies)

    return resources


###############################################################################
# Private functions
###############################################################################
async def explore_network_policies(c: AsyncClient) -> list[Resource]:
    response = await c.execute("list_network_policy_for_all_namespaces")

    policies = msgspec.json.decode(response.data)

    results = []
    for policy in policies["items"]:
        meta = policy["metadata"]
        results.append(
            Resource(
                id=make_id(meta["uid"]),
                meta=K8SMeta(
                    name=meta["name"],
                    display=meta["name"],
                    kind="network-policy",
                    platform="k8s",
                    namespace=meta.get("namespace"),
                    category="security",
                ),
                struct=policy,
            )
        )

    return results



# mypy: disable-error-code="attr-defined,operator,union-attr"
import asyncio
from typing import Any

import msgspec
from google.oauth2._service_account_async import Credentials

from lueur.links import add_link
from lueur.make_id import make_id
from lueur.models import Discovery, GCPMeta, Link, Resource
from lueur.platform.gcp.client import AuthorizedSession, Client
from lueur.rules import iter_resource

__all__ = ["explore_monitoring"]


async def explore_monitoring(
    project: str, creds: Credentials | None = None
) -> list[Resource]:
    resources = []

    async with Client("https://monitoring.googleapis.com", creds) as c:
        services = await explore_services(c, project)
        resources.extend(services)

        tasks: list[asyncio.Task] = []

        async with asyncio.TaskGroup() as tg:
            tasks.append(tg.create_task(explore_alert_policies(c, project)))
            tasks.append(tg.create_task(explore_groups(c, project)))
            tasks.append(
                tg.create_task(explore_notification_channels(c, project))
            )

            # uncomment once API is publicly available
            # tasks.append(tg.create_task(explore_incidents(c, project)))

            for svc in services:
                tasks.append(
                    tg.create_task(explore_slos(c, project, svc.meta.name))
                )

        for task in tasks:
            result = task.result()
            if result is None:
                continue
            resources.extend(result)

    return resources


###############################################################################
# Private functions
###############################################################################
async def explore_services(
    c: AuthorizedSession, project: str
) -> list[Resource]:
    response = await c.get(f"/v3/projects/{project}/services")

    services = msgspec.json.decode(response.content)

    results = []
    for svc in services["services"]:
        name = svc["name"]
        display = svc["displayName"]
        self_link = svc.get("selfLink")

        results.append(
            Resource(
                id=make_id(name),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="service",
                    project=project,
                    platform="gcp",
                    self_link=self_link,
                    category="observability",
                ),
                struct=svc,
            )
        )

    return results


async def explore_slos(
    c: AuthorizedSession, project: str, service_name: str
) -> list[Resource]:
    response = await c.get(
        f"/v3/{service_name}/serviceLevelObjectives",
        params={"view": "EXPLICIT"},
    )

    slos = msgspec.json.decode(response.content)

    if not slos:
        return []

    results = []
    for slo in slos["serviceLevelObjectives"]:
        name = slo["name"]
        display = slo["displayName"]
        self_link = slo.get("selfLink")

        results.append(
            Resource(
                id=make_id(slo["name"]),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="slo",
                    project=project,
                    platform="gcp",
                    self_link=self_link,
                    category="observability",
                ),
                struct=slo,
            )
        )

    return results


async def explore_alert_policies(
    c: AuthorizedSession, project: str
) -> list[Resource]:
    response = await c.get(f"/v3/projects/{project}/alertPolicies")

    policies = msgspec.json.decode(response.content)

    results = []
    for nc in policies["alertPolicies"]:
        name = nc["name"]
        display = nc["displayName"]
        self_link = nc.get("selfLink")

        results.append(
            Resource(
                id=make_id(name),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="alert-policy",
                    project=project,
                    platform="gcp",
                    self_link=self_link,
                    category="observability",
                ),
                struct=nc,
            )
        )

    return results


async def explore_notification_channels(
    c: AuthorizedSession, project: str
) -> list[Resource]:
    response = await c.get(f"/v3/projects/{project}/notificationChannels")

    channels = msgspec.json.decode(response.content)

    results = []
    for nc in channels["notificationChannels"]:
        name = nc["name"]
        display = nc["displayName"]
        self_link = nc.get("selfLink")

        results.append(
            Resource(
                id=make_id(name),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="notification-channel",
                    project=project,
                    platform="gcp",
                    self_link=self_link,
                    category="observability",
                ),
                struct=nc,
            )
        )

    return results


async def explore_groups(c: AuthorizedSession, project: str) -> list[Resource]:
    response = await c.get(f"/v3/projects/{project}/groups")

    groups = msgspec.json.decode(response.content)

    results = []
    for g in groups.get("group", []):
        name = g["name"]
        display = g["displayName"]
        self_link = g.get("selfLink")

        results.append(
            Resource(
                id=make_id(name),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="group",
                    project=project,
                    platform="gcp",
                    self_link=self_link,
                    category="observability",
                ),
                struct=g,
            )
        )

    return results


async def explore_incidents(
    c: AuthorizedSession, project: str
) -> list[Resource]:
    response = await c.get(f"/v3/projects/{project}/incidents")

    incidents = msgspec.json.decode(response.content)

    results = []
    for g in incidents.get("incidents", []):
        name = g["name"]
        display = g["name"]
        self_link = g.get("selfLink")

        results.append(
            Resource(
                id=make_id(name),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="incident",
                    project=project,
                    platform="gcp",
                    self_link=self_link,
                    category="observability",
                ),
                struct=g,
            )
        )

    return results


def expand_links(d: Discovery, serialized: dict[str, Any]) -> None:
    for s in iter_resource(
        serialized, "$.resources[?@.meta.kind=='service'].meta.name"
    ):
        r_id = s.parent.parent.obj["id"]  # type: ignore
        name = s.value
        p = f"$.resources[?@.meta.kind=='slo' && match(@.meta.name, '{name}/serviceLevelObjectives/.*')]"  # noqa E501
        for slo in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="slo",
                    path=slo.path,
                    pointer=str(slo.pointer()),
                ),
            )

    for slo in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='slo'].meta.name",
    ):
        r_id = slo.parent.parent.obj["id"]  # type: ignore
        segments = slo.value.split("/")
        name = "/".join(segments[:-2])  # service name

        p = f"$.resources[?@.meta.kind=='service' && @.meta.name=='{name}']"
        for svc in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="service",
                    path=svc.path,
                    pointer=str(svc.pointer()),
                ),
            )

    for alertnc in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='alert-policy'].struct.notificationChannels.*",  # noqa E501
    ):
        r_id = alertnc.parent.parent.parent.obj["id"]  # type: ignore
        name = alertnc.value
        p = f"$.resources[?@.meta.kind=='notification-channel' && @.meta.name=='{name}']"  # noqa E501
        for nc in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="notification-channel",
                    path=nc.path,
                    pointer=str(nc.pointer()),
                ),
            )

    slos = {}
    for slo in iter_resource(
        serialized, "$.resources[?@.meta.kind=='slo'].meta.name"
    ):
        r_id = slo.parent.parent.obj["id"]  # type: ignore
        slos[slo.value] = r_id

    for condition_filter in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='alert-policy'].struct.conditions.*.conditionThreshold.filter",  # noqa E501
    ):
        cf = condition_filter.value
        for slo in slos:  # type: ignore
            if slo in cf:
                add_link(
                    d,
                    slos[slo],
                    Link(
                        direction="out",
                        kind="alert-policy",
                        path=condition_filter.path,
                        pointer=str(condition_filter.pointer()),
                    ),
                )
                break

    for s in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='service' && @.struct.basicService.serviceType=='CLOUD_RUN'].struct.basicService.serviceLabels.service_name",  # noqa E501
    ):
        r_id = s.parent.parent.parent.parent.obj["id"]  # type: ignore
        name = s.value
        p = f"$.resources[?@.meta.kind=='global-backend-service' && @.meta.name=='{name}']"  # noqa E501
        for slo in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="global-backend-service",
                    path=slo.path,
                    pointer=str(slo.pointer()),
                ),
            )



from datetime import datetime
from typing import Any, Literal, cast

import msgspec
from pydantic import BaseModel, Field, RootModel

from lueur.date_time import now

__all__ = [
    "Resource",
    "Meta",
    "Discovery",
    "Recommendation",
    "Rule",
    "Rules",
    "GCPProject",
    "K8SMeta",
    "AWSMeta",
]


class Link(BaseModel):
    direction: Literal["in", "out"]
    kind: str
    path: str
    pointer: str


class Meta(BaseModel):
    name: str
    display: str
    dt: datetime = Field(default_factory=now)
    kind: str
    platform: str | None = None
    category: (
        Literal[
            "network",
            "observability",
            "storage",
            "compute",
            "security",
            "loadbalancer",
            "integration",
            "memorystore",
            "sql",
            "gke"
        ]
        | None
    )


class GCPMeta(Meta):
    project: str
    region: str | None = None
    zone: str | None = None
    self_link: str | None = None
    regional: bool = False


class K8SMeta(Meta):
    ns: str | None = None


class AWSMeta(Meta):
    region: str | None = None


class Resource(BaseModel):
    id: str
    meta: Meta | GCPMeta | AWSMeta | K8SMeta
    links: list[Link] = Field(default_factory=list)
    struct: dict


class Discovery(BaseModel):
    id: str
    meta: Meta
    resources: list[Resource]


class Recommendation(BaseModel):
    id: str
    meta: Meta


class JSonPathRuleMatcher(BaseModel):
    type: Literal["jsonpath"]
    path: str


class Rule(BaseModel):
    key: str
    target_kinds: list[str]
    reason: str
    explanation: str
    severity: Literal["critical", "warning", "info"]
    matcher: JSonPathRuleMatcher


Rules = RootModel[list[Rule]]


class FaultScenarioVar(BaseModel):
    type: Literal["env"] = "env"
    key: str
    default: bytes | dict | list | int | float | str | bool | None = None
    env_var_type: (
        Literal["bytes", "int", "float", "bool", "str", "json"] | None
    ) = "str"


Activity = RootModel[dict[str, Any]]


class FaultScenario(BaseModel):
    title: str
    purpose: str
    tags: list[str]
    vars: dict[str, FaultScenarioVar]
    faults: list[Activity]
    verification: dict[str, str | list[Activity]] | None = None
    rollbacks: list[Activity] | None = None

    def as_experiment(self) -> dict[str, Any]:
        x = {
            "title": self.title,
            "description": self.purpose,
            "tags": self.tags,
            "runtime": {
                "hypothesis": {"strategy": "default"},
                "rollbacks": {"strategy": "always"},
            },
            "configuration": {k: v.model_dump() for k, v in self.vars.items()},
            "method": [],
        }

        if self.verification:
            probes = cast(list[Activity], self.verification["probes"])
            x["steady-state-hypothesis"] = {
                "title": self.verification["title"],
                "probes": [f.model_dump() for f in probes],
            }

        if self.faults:
            x["method"] = [f.model_dump() for f in self.faults]

        if self.rollbacks:
            x["rollbacks"] = [f.model_dump() for f in self.rollbacks]

        return x


class GCPProject(BaseModel):
    id: str
    name: str
    organization: str


class SnapshotMeta(msgspec.Struct):
    name: str
    display: str
    kind: str
    region: str | None = None
    zone: str | None = None


class SnapshotLink(msgspec.Struct):
    pointer: str


class SnapshotResource(msgspec.Struct):
    meta: SnapshotMeta
    links: list[SnapshotLink]


class Snapshot(msgspec.Struct):
    resources: list[SnapshotResource]


class AbortFaultPolicy(BaseModel):
    http_status: int
    percentage: float


class DelayFaultPolicy(BaseModel):
    fixed_delay_seconds: str
    fixed_delay_nanos: int
    percentage: float


class HostRedirect(BaseModel):
    host: str
    path: str


class HostPath(BaseModel):
    paths: list[str]
    faults: list[AbortFaultPolicy | DelayFaultPolicy] | None = None


class HostRouteRule(BaseModel):
    priority: int
    kind: str
    path: str


class HostRoute(BaseModel):
    rules: list[HostRouteRule]


class HostRouteAction(BaseModel):
    default: bool = False


class Host(BaseModel):
    domains: list[str]
    paths: HostPath | HostRoute | HostRedirect | HostRouteAction


class LoadBalancer(BaseModel):
    name: str
    display: str
    region: str | None = None
    zone: str | None = None
    hosts: list[Host]
    snapshot_pointer: str


class SiteEntities(BaseModel):
    lb: list[LoadBalancer]


class Site(BaseModel):
    entities: SiteEntities


class ChatMessage(msgspec.Struct):
    role: Literal["system", "user"]
    content: str


class LatencyChat(msgspec.Struct):
    prelude: list[ChatMessage]
    ready: list[ChatMessage]
    postface: list[ChatMessage]


class HTTPErrorChat(msgspec.Struct):
    prelude: list[ChatMessage]
    ready: list[ChatMessage]
    postface: list[ChatMessage]


class CTKChat(msgspec.Struct):
    postface: list[ChatMessage]



import msgspec
from google.oauth2._service_account_async import Credentials

from lueur.platform.gcp.client import Client

__all__ = ["list_project_zones", "list_project_region_zones"]


async def list_project_zones(
    project: str, creds: Credentials | None = None
) -> list[str]:
    async with Client("https://compute.googleapis.com", creds) as c:
        response = await c.get(
            f"/compute/v1/projects/{project}/zones",
            params={"fields": "items.name"},
        )

        zones = msgspec.json.decode(response.content)

        return [r["name"] for r in zones["items"]]


async def list_project_region_zones(
    project: str, region: str, creds: Credentials | None = None
) -> list[str]:
    async with Client("https://compute.googleapis.com", creds) as c:
        response = await c.get(
            f"/compute/v1/projects/{project}/regions/{region}/zones",
            params={"fields": "items.name"},
        )

        zones = msgspec.json.decode(response.content)

        return [r["name"] for r in zones.get("items", [])]



# mypy: disable-error-code="union-attr"
import asyncio
from typing import Any

import msgspec
from google.oauth2._service_account_async import Credentials

from lueur.links import add_link
from lueur.make_id import make_id
from lueur.models import Discovery, GCPMeta, Link, Resource
from lueur.platform.gcp.client import AuthorizedSession, Client
from lueur.rules import iter_resource

__all__ = ["explore_gke"]


async def explore_gke(
    project: str, location: str, creds: Credentials | None = None
) -> list[Resource]:
    resources = []

    async with Client("https://container.googleapis.com", creds) as c:
        clusters = await explore_clusters(c, project, location)
        resources.extend(clusters)

        tasks: list[asyncio.Task] = []

        async with asyncio.TaskGroup() as tg:
            for cl in clusters:
                tasks.append(
                    tg.create_task(
                        explore_nodepools(c, project, location, cl.meta.name)
                    )
                )

        for task in tasks:
            result = task.result()
            if result is None:
                continue
            resources.extend(result)

    return resources


###############################################################################
# Private functions
###############################################################################
async def explore_clusters(
    c: AuthorizedSession, project: str, location: str
) -> list[Resource]:
    response = await c.get(
        f"/v1/projects/{project}/locations/{location}/clusters"
    )

    clusters = msgspec.json.decode(response.content)

    results = []
    for cl in clusters["clusters"]:
        self_link = cl.get("selfLink")

        results.append(
            Resource(
                id=make_id(cl["id"]),
                meta=GCPMeta(
                    name=cl["name"],
                    display=cl["name"],
                    kind="cluster",
                    platform="gcp",
                    category="gke",
                    project=project,
                    region=location,
                    self_link=self_link,
                ),
                struct=cl,
            )
        )

    return results


async def explore_nodepools(
    c: AuthorizedSession, project: str, location: str, cluster_name: str
) -> list[Resource]:
    response = await c.get(
        f"/v1/projects/{project}/locations/{location}/clusters/{cluster_name}/nodePools"  # noqa E501
    )

    node_pools = msgspec.json.decode(response.content)

    results = []
    for np in node_pools["nodePools"]:
        self_link = np.get("selfLink")

        results.append(
            Resource(
                id=make_id(np["selfLink"]),
                meta=GCPMeta(
                    name=np["name"],
                    display=np["name"],
                    kind="nodepool",
                    category="gke",
                    platform="gcp",
                    project=project,
                    region=location,
                    self_link=self_link,
                ),
                struct=np,
            )
        )

    return results


def expand_links(d: Discovery, serialized: dict[str, Any]) -> None:
    for np in iter_resource(
        serialized, "$.resources[?@.meta.kind=='cluster'].struct.network"
    ):
        r_id = np.parent.parent.obj["id"]  # type: ignore
        name = np.value
        p = f"$.resources[?@.meta.kind=='k8s/node' && @.struct.metadata.labels.['cloud.google.com/gke-nodepool']=='{name}']"  # noqa E501
        for k8snode in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="k8s/node",
                    path=k8snode.path,
                    pointer=str(k8snode.pointer()),
                ),
            )

        p = f"$.resources[?@.meta.kind=='network' && @.meta.name=='{name}']"
        for k8snode in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="network",
                    path=k8snode.path,
                    pointer=str(k8snode.pointer()),
                ),
            )

    for np in iter_resource(
        serialized, "$.resources[?@.meta.kind=='cluster'].struct.subnetwork"
    ):
        p = f"$.resources[?@.meta.kind=='subnet' && @.meta.name=='{name}']"
        for k8snode in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="subnet",
                    path=k8snode.path,
                    pointer=str(k8snode.pointer()),
                ),
            )

    for np in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='cluster'].struct.nodePools.*.name",
    ):
        r_id = np.parent.parent.parent.parent.obj["id"]  # type: ignore
        np_name = np.value
        p = f"$.resources[?@.meta.kind=='nodepool' && @.meta.name=='{np_name}']"
        for svc in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="nodepool",
                    path=svc.path,
                    pointer=str(svc.pointer()),
                ),
            )

    for s in iter_resource(
        serialized, "$.resources[?@.meta.kind=='cluster'].meta.name"
    ):
        r_id = s.parent.parent.obj["id"]  # type: ignore
        name = s.value
        p = f"$.resources[?@.meta.kind=='service' && @.struct.gkeWorkload.clusterName=='{name}']"  # noqa E501
        for service in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="service",
                    path=service.path,
                    pointer=str(service.pointer()),
                ),
            )



# mypy: disable-error-code="index,call-arg,union-attr"
from typing import Any

import msgspec
from kubernetes import client

from lueur.links import add_link
from lueur.make_id import make_id
from lueur.models import Discovery, K8SMeta, Link, Resource
from lueur.platform.k8s.client import AsyncClient, Client
from lueur.rules import iter_resource

__all__ = ["explore_node"]


async def explore_node() -> list[Resource]:
    resources = []

    async with Client(client.CoreV1Api) as c:
        nodes = await explore_nodes(c)
        resources.extend(nodes)

    return resources


###############################################################################
# Private functions
###############################################################################
async def explore_nodes(c: AsyncClient) -> list[Resource]:
    response = await c.execute("list_node")

    nodes = msgspec.json.decode(response.data)

    results = []
    for node in nodes["items"]:
        meta = node["metadata"]
        results.append(
            Resource(
                id=make_id(meta["uid"]),
                meta=K8SMeta(
                    name=meta["name"],
                    display=meta["name"],
                    kind="node",
                    platform="k8s",
                    category="compute",
                ),
                struct=node,
            )
        )

    return results


def expand_links(d: Discovery, serialized: dict[str, Any]) -> None:
    for np in iter_resource(
        serialized, "$.resources[?@.meta.kind=='nodepool'].meta.name"
    ):
        r_id = np.parent.parent.obj["id"]
        name = np.value
        p = f"$.resources[?@.meta.kind=='node' && @.struct.metadata.labels.['cloud.google.com/gke-nodepool']=='{name}']"  # noqa E501
        for k8snode in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="node",
                    path=k8snode.path,
                    pointer=str(k8snode.pointer()),
                ),
            )



# mypy: disable-error-code="arg-type,attr-defined,index,union-attr"
import asyncio
from typing import Any, cast

import msgspec
from google.oauth2._service_account_async import Credentials

from lueur.links import add_link
from lueur.make_id import make_id
from lueur.models import Discovery, GCPMeta, Link, Resource
from lueur.platform.gcp.client import AuthorizedSession, Client
from lueur.platform.gcp.zone import list_project_zones
from lueur.rules import iter_resource

__all__ = ["explore_lb"]


async def explore_lb(
    project: str, location: str | None = None, creds: Credentials | None = None
) -> list[Resource]:
    resources = []

    async with Client("https://compute.googleapis.com", creds) as c:
        tasks: list[asyncio.Task] = []

        async with asyncio.TaskGroup() as tg:
            if not location:
                tasks.append(tg.create_task(explore_global_urlmaps(c, project)))
                tasks.append(
                    tg.create_task(explore_global_backend_services(c, project))
                )
                tasks.append(
                    tg.create_task(explore_global_backend_groups(c, project))
                )
                tasks.append(
                    tg.create_task(explore_zonal_backend_groups(c, project))
                )
            else:
                tasks.append(
                    tg.create_task(
                        explore_regional_urlmaps(c, project, location)
                    )
                )
                tasks.append(
                    tg.create_task(
                        explore_regional_backend_services(c, project, location)
                    )
                )
                tasks.append(
                    tg.create_task(
                        explore_regional_backend_groups(c, project, location)
                    )
                )

        for task in tasks:
            result = task.result()
            if result is None:
                continue
            resources.extend(result)

    return resources


###############################################################################
# Private functions
###############################################################################
async def explore_global_urlmaps(
    c: AuthorizedSession, project: str
) -> list[Resource]:
    response = await c.get(
        f"/compute/v1/projects/{project}/global/urlMaps",
        params={"returnPartialSuccess": True},
    )

    urlmaps = msgspec.json.decode(response.content)

    results = []
    for urlmap in urlmaps.get("items", []):
        results.append(
            Resource(
                id=make_id(urlmap["id"]),
                meta=GCPMeta(
                    name=urlmap["name"],
                    display=urlmap["name"],
                    kind="global-urlmap",
                    project=project,
                    platform="gcp",
                    category="loadbalancer",
                ),
                struct=urlmap,
            )
        )

    return results


async def explore_global_backend_services(
    c: AuthorizedSession, project: str
) -> list[Resource]:
    response = await c.get(
        f"/compute/v1/projects/{project}/global/backendServices",
        params={"returnPartialSuccess": True},
    )

    backend_services = msgspec.json.decode(response.content)

    results = []
    for backend_service in backend_services.get("items", []):
        name = backend_service["name"]
        display = name
        self_link = backend_service.get("selfLink")

        results.append(
            Resource(
                id=make_id(backend_service["id"]),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="global-backend-service",
                    project=project,
                    platform="gcp",
                    self_link=self_link,
                    category="loadbalancer",
                ),
                struct=backend_service,
            )
        )

    return results


async def explore_global_backend_groups(
    c: AuthorizedSession, project: str
) -> list[Resource]:
    response = await c.get(
        f"/compute/v1/projects/{project}/global/networkEndpointGroups",
        params={"returnPartialSuccess": True},
    )

    backend_groups = msgspec.json.decode(response.content)

    results = []
    for backend_group in backend_groups.get("items", []):
        name = backend_group["name"]
        display = name
        region = backend_group.get("region")
        zone = backend_group.get("zone")
        self_link = backend_group.get("selfLink")
        if zone:
            _, zone = zone.rsplit("/", 1)

        results.append(
            Resource(
                id=make_id(backend_group["id"]),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="global-neg",
                    project=project,
                    region=region,
                    zone=zone,
                    platform="gcp",
                    self_link=self_link,
                    category="loadbalancer",
                ),
                struct=backend_group,
            )
        )

    return results


async def explore_regional_urlmaps(
    c: AuthorizedSession, project: str, location: str
) -> list[Resource]:
    response = await c.get(
        f"/compute/v1/projects/{project}/regions/{location}/urlMaps",
        params={"returnPartialSuccess": True},
    )

    urlmaps = msgspec.json.decode(response.content)

    results = []
    for urlmap in urlmaps.get("items", []):
        self_link = urlmap.get("selfLink")

        results.append(
            Resource(
                id=make_id(urlmap["id"]),
                meta=GCPMeta(
                    name=urlmap["name"],
                    display=urlmap["name"],
                    kind="regional-urlmap",
                    project=project,
                    region=location,
                    platform="gcp",
                    self_link=self_link,
                    category="loadbalancer",
                ),
                struct=urlmap,
            )
        )

    return results


async def explore_regional_backend_services(
    c: AuthorizedSession, project: str, location: str
) -> list[Resource]:
    response = await c.get(
        f"/compute/v1/projects/{project}/regions/{location}/backendServices",
        params={"returnPartialSuccess": True},
    )

    backend_services = msgspec.json.decode(response.content)

    results = []
    for backend_service in backend_services.get("items", []):
        name = backend_service["name"]
        region = backend_service["region"]
        self_link = backend_service.get("selfLink")
        display = name

        results.append(
            Resource(
                id=make_id(backend_service["id"]),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="regional-backend-service",
                    project=project,
                    region=region,
                    platform="gcp",
                    self_link=self_link,
                    category="loadbalancer",
                ),
                struct=backend_service,
            )
        )

    return results


async def explore_regional_backend_groups(
    c: AuthorizedSession, project: str, location: str
) -> list[Resource]:
    response = await c.get(
        f"/compute/v1/projects/{project}/regions/{location}/networkEndpointGroups",
        params={"returnPartialSuccess": True},
    )

    backend_groups = msgspec.json.decode(response.content)

    results = []

    for backend_group in backend_groups.get("items", []):
        name = backend_group["name"]
        display = name
        self_link = backend_group.get("selfLink")
        zone = backend_group.get("zone")
        if zone:
            _, zone = zone.rsplit("/", 1)

        results.append(
            Resource(
                id=make_id(backend_group["id"]),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="regional-neg",
                    project=project,
                    region=location,
                    zone=zone,
                    platform="gcp",
                    self_link=self_link,
                    category="loadbalancer",
                ),
                struct=backend_group,
            )
        )

    return results


async def explore_zonal_backend_groups(
    c: AuthorizedSession, project: str
) -> list[Resource]:
    zones = await list_project_zones(project)

    results = []
    tasks: list[asyncio.Task] = []

    async with asyncio.TaskGroup() as tg:
        for zone in zones:
            tasks.append(
                tg.create_task(
                    c.get(
                        f"/compute/v1/projects/{project}/zones/{zone}/networkEndpointGroups",
                        params={"returnPartialSuccess": True},
                    )
                )
            )

    for task in tasks:
        response = task.result()
        if response is None:
            response

        backend_groups = msgspec.json.decode(response.content)

        for backend_group in backend_groups.get("items", []):
            name = backend_group["name"]
            display = name
            self_link = backend_group.get("selfLink")
            region = backend_group.get("region")
            if region:
                _, region = region.rsplit("/", 1)

            results.append(
                Resource(
                    id=make_id(backend_group["id"]),
                    meta=GCPMeta(
                        name=name,
                        display=display,
                        kind="zonal-neg",
                        project=project,
                        region=region,
                        zone=zone,
                        platform="gcp",
                        self_link=self_link,
                        category="loadbalancer",
                    ),
                    struct=backend_group,
                )
            )

    return results


def expand_links(d: Discovery, serialized: dict[str, Any]) -> None:
    for backend_service in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='global-backend-service'].meta.name",
    ):
        struct = backend_service.parent.parent.obj["struct"]  # type: ignore
        for used_by in struct.get("usedBy", []):
            ref = used_by.get("reference")
            p = f"$.resources[?@.meta.kind=='global-urlmap' && @.struct.selfLink=='{ref}'].id"  # noqa E501
            for urlmap_id in iter_resource(serialized, p):
                add_link(
                    d,
                    urlmap_id.value,
                    Link(
                        direction="out",
                        kind="global-backend-service",
                        path=backend_service.parent.parent.path,  # type: ignore
                        pointer=str(backend_service.parent.parent.pointer()),  # type: ignore
                    ),
                )

    for s in iter_resource(
        serialized, "$.resources[?@.meta.kind=='global-neg'].struct.subnetwork"
    ):
        r_id = s.parent.parent.obj["id"]  # type: ignore
        subnet = s.value
        p = f"$.resources[?@.meta.kind=='subnet' && @.meta.self_link=='{subnet}']"  # noqa E501
        for subnet in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="subnet",
                    path=subnet.path,
                    pointer=str(subnet.pointer()),
                ),
            )

    for s in iter_resource(
        serialized, "$.resources[?@.meta.kind=='zonal-neg'].struct.network"
    ):
        r_id = s.parent.parent.obj["id"]  # type: ignore
        network = s.value
        p = f"$.resources[?@.meta.kind=='network' && @.meta.self_link=='{network}']"  # noqa E501
        for subnet in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="network",
                    path=subnet.path,
                    pointer=str(subnet.pointer()),
                ),
            )

    for s in iter_resource(
        serialized, "$.resources[?@.meta.kind=='zonal-neg'].struct.subnetwork"
    ):
        r_id = s.parent.parent.obj["id"]  # type: ignore
        subnet = s.value
        p = f"$.resources[?@.meta.kind=='subnet' && @.meta.self_link=='{subnet}']"  # noqa E501
        for subnet in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="subnet",
                    path=subnet.path,
                    pointer=str(subnet.pointer()),
                ),
            )

    for s in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='global-backend-service'].struct.backends.*.group",  # noqa E501
    ):
        r_id = s.parent.parent.parent.parent.obj["id"]  # type: ignore
        group = s.value
        p = f"$.resources[?@.meta.kind=='regional-neg' && @.struct.selfLink=='{group}'].struct.cloudRun.service"  # noqa E501
        for service in iter_resource(serialized, p):
            neg = service.parent.parent.parent.obj["meta"]  # type: ignore
            project = neg["project"]
            location = neg["region"]

            cloudrun_name = f"projects/{project}/locations/{location}/services/{service.obj}"  # noqa E501
            p = f"$.resources[?@.meta.kind=='cloudrun' && @.meta.name=='{cloudrun_name}']"  # noqa E501
            for subnet in iter_resource(serialized, p):
                add_link(
                    d,
                    r_id,
                    Link(
                        direction="out",
                        kind="subnet",
                        path=subnet.path,
                        pointer=str(subnet.pointer()),
                    ),
                )

        p = f"$.resources[?@.meta.kind=='zonal-neg' && @.struct.selfLink=='{group}']"  # noqa E501
        for zonal_neg in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="zonal-neg",
                    path=zonal_neg.path,
                    pointer=str(zonal_neg.pointer()),
                ),
            )

    for s in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='global-backend-service'].struct.securityPolicy",  # noqa E501
    ):
        r_id = s.parent.parent.obj["id"]  # type: ignore
        secpolicy = s.value
        p = f"$.resources[?@.meta.kind=='global-securities' && @.meta.self_link=='{secpolicy}']"  # noqa E501
        for service in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="global-securities",
                    path=service.path,
                    pointer=str(service.pointer()),
                ),
            )

        p = f"$.resources[?@.meta.kind=='regional-securities' && @.meta.self_link=='{secpolicy}']"  # noqa E501
        for service in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="regional-securities",
                    path=service.path,
                    pointer=str(service.pointer()),
                ),
            )

    for s in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='regional-backend-service'].struct.securityPolicy",  # noqa E501
    ):
        r_id = cast(str, s.parent.parent.obj["id"])  # type: ignore
        secpolicy = s.value
        p = f"$.resources[?@.meta.kind=='regional-securities' && @.meta.self_link=='{secpolicy}']"  # noqa E501
        for service in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="regional-securities",
                    path=service.path,
                    pointer=str(service.pointer()),
                ),
            )

    for s in iter_resource(
        serialized, "$.resources[?@.meta.kind=='regional-neg'].struct.network"
    ):
        r_id = cast(str, s.parent.parent.obj["id"])  # type: ignore
        network = s.value
        p = f"$.resources[?@.meta.kind=='network' && @.meta.self_link=='{network}']"  # noqa E501
        for subnet in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="network",
                    path=subnet.path,
                    pointer=str(subnet.pointer()),
                ),
            )

    for svcneg in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='global-backend-service'].struct.backends.*.group",
    ):
        r_id = svcneg.parent.parent.parent.parent.obj["id"]  # type: ignore
        neg = svcneg.value
        p = f"$.resources[?@.meta.kind=='regional-neg' && @.struct.selfLink=='{neg}'].struct.cloudRun.service"  # noqa E501
        for cloudrun_name in iter_resource(serialized, p):  # type: ignore
            svc_name = cloudrun_name.value
            p = f"$.resources[?@.meta.kind=='cloudrun' && @.meta.name contains '{svc_name}']"  # noqa E501
            for cloudrun in iter_resource(serialized, p):
                add_link(
                    d,
                    r_id,
                    Link(
                        direction="out",
                        kind="cloudrun",
                        path=cloudrun.path,
                        pointer=str(cloudrun.pointer()),
                    ),
                )

                p = f"$.resources[?@.meta.kind=='service' && @.struct.cloudRun.serviceName=='{svc_name}']"  # noqa E501
                for svc in iter_resource(serialized, p):
                    add_link(
                        d,
                        r_id,
                        Link(
                            direction="out",
                            kind="service",
                            path=svc.path,
                            pointer=str(svc.pointer()),
                        ),
                    )

                    full_svc_name = svc.obj["meta"]["name"]  # type: ignore

                    p = f"$.resources[?@.meta.kind=='slo' && match(@.meta.name, '{full_svc_name}/serviceLevelObjectives/.*')]"  # noqa E501
                    for slo in iter_resource(serialized, p):
                        add_link(
                            d,
                            r_id,
                            Link(
                                direction="out",
                                kind="slo",
                                path=slo.path,
                                pointer=str(slo.pointer()),
                            ),
                        )

    for urlmap in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='gateway'].struct.metadata.annotations['networking.gke.io/url-maps']",  # noqa E501
    ):
        r_id = urlmap.parent.parent.parent.parent.obj["id"]  # type: ignore
        name = urlmap.value.rsplit("/", 1)[-1]  # type: ignore

        p = f"$.resources[?@.meta.kind=='global-urlmap' && @.meta.name=='{name}']"  # noqa E501
        for urlmap in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="global-urlmap",
                    path=urlmap.path,
                    pointer=str(urlmap.pointer()),
                ),
            )



# mypy: disable-error-code="index,union-attr"
import asyncio
from typing import Any

import msgspec
from google.oauth2._service_account_async import Credentials

from lueur.links import add_link
from lueur.make_id import make_id
from lueur.models import Discovery, GCPMeta, Link, Resource
from lueur.platform.gcp.client import AuthorizedSession, Client
from lueur.rules import iter_resource

__all__ = ["explore_sql", "expand_links"]


async def explore_sql(
    project: str, creds: Credentials | None = None
) -> list[Resource]:
    resources = []

    async with Client("https://sqladmin.googleapis.com", creds) as c:
        instances = await explore_instances(c, project)
        resources.extend(instances)

        tasks: list[asyncio.Task] = []

        async with asyncio.TaskGroup() as tg:
            for inst in instances:
                tasks.append(
                    tg.create_task(explore_users(c, project, inst.meta.name))
                )
                tasks.append(
                    tg.create_task(
                        explore_databases(c, project, inst.meta.name)
                    )
                )

        for task in tasks:
            result = task.result()
            if result is None:
                continue
            resources.extend(result)

    return resources


###############################################################################
# Private functions
###############################################################################
async def explore_instances(
    c: AuthorizedSession, project: str
) -> list[Resource]:
    response = await c.get(f"/v1/projects/{project}/instances")

    instances = msgspec.json.decode(response.content)

    results = []
    for inst in instances["items"]:
        self_link = inst.get("selfLink")

        results.append(
            Resource(
                id=make_id(inst["selfLink"]),
                meta=GCPMeta(
                    name=inst["name"],
                    display=inst["name"],
                    kind="instance",
                    platform="gcp",
                    project=project,
                    self_link=self_link,
                    category="sql",
                ),
                struct=inst,
            )
        )

    return results


async def explore_users(
    c: AuthorizedSession, project: str, instance: str
) -> list[Resource]:
    response = await c.get(f"/v1/projects/{project}/instances/{instance}/users")

    users = msgspec.json.decode(response.content)

    results = []
    for user in users["items"]:
        self_link = user.get("selfLink")

        results.append(
            Resource(
                id=make_id(f"{user['instance']}-{user['name']}"),
                meta=GCPMeta(
                    name=user["name"],
                    display=user["name"],
                    kind="user",
                    platform="gcp",
                    project=project,
                    self_link=self_link,
                    category="sql",
                ),
                struct=user,
            )
        )

    return results


async def explore_databases(
    c: AuthorizedSession, project: str, instance: str
) -> list[Resource]:
    response = await c.get(
        f"/v1/projects/{project}/instances/{instance}/databases"
    )

    databases = msgspec.json.decode(response.content)

    results = []
    for database in databases["items"]:
        self_link = database.get("selfLink")

        results.append(
            Resource(
                id=make_id(f"{database['instance']}-{database['name']}"),
                meta=GCPMeta(
                    name=database["name"],
                    display=database["name"],
                    kind="database",
                    platform="gcp",
                    project=project,
                    self_link=self_link,
                    category="sql",
                ),
                struct=database,
            )
        )

    return results


def expand_links(d: Discovery, serialized: dict[str, Any]) -> None:
    for i in iter_resource(
        serialized, "$.resources[?@.meta.kind=='instance'].meta.name"
    ):
        r_id = i.parent.parent.obj["id"]
        name = i.value

        p = f"$.resources[?@.meta.kind=='user' && @.struct.instance=='{name}']"
        for user in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="user",
                    path=user.path,
                    pointer=str(user.pointer()),
                ),
            )

        p = f"$.resources[?@.meta.kind=='database' && @.struct.instance=='{name}']"  # noqa E501
        for db in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="database",
                    path=db.path,
                    pointer=str(db.pointer()),
                ),
            )

    for s in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='instance'].struct.settings.ipConfiguration.privateNetwork",  # noqa E501
    ):
        r_id = s.parent.parent.parent.parent.obj["id"]  # type: ignore
        network = s.value.rsplit("/", 1)[-1]  # type: ignore
        p = f"$.resources[?@.meta.kind=='network' && @.meta.name=='{network}']"
        for svc in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="network",
                    path=svc.path,
                    pointer=str(svc.pointer()),
                ),
            )



import secrets
from datetime import datetime, timezone
from pathlib import Path

from .make_id import make_id
from .models import Discovery, Meta, Resource

__all__ = ["merge_discoveries", "load_discovery"]


def merge_discoveries(
    *discoveries: Discovery,
) -> Discovery:
    name = secrets.token_hex(8)

    resources: list[Resource] = []
    for d in discoveries:
        resources.extend(d.resources)

    return Discovery(
        id=make_id(name),
        resources=resources,
        meta=Meta(
            name=name,
            display=f"Snaphot {datetime.now(timezone.utc)}",
            kind="mixed",
            category=None,
        ),
    )


def load_discovery(snapshot_file: Path) -> Discovery:
    return Discovery.model_validate_json(snapshot_file.read_bytes())



from typing import Any

import msgspec
import pytest
import respx

from lueur.platform.gcp.sql import explore_sql

from tests.platform.gcp.utils import patch_auth


@respx.mock
@pytest.mark.asyncio
async def test_explore_sql(
    gcp_project: str,
    gcp_sql_instance_name: str,
    gcp_sql_instances: list[dict[str, Any]],
    gcp_sql_users: list[dict[str, Any]],
    gcp_sql_databases: list[dict[str, Any]],
) -> None:
    respx.post(
        "https://oauth2.googleapis.com/token",
    ).respond(content=msgspec.json.encode({"access_token": "xyz"}))

    respx.get(
        f"https://sqladmin.googleapis.com/v1/projects/{gcp_project}/instances",
    ).respond(content=msgspec.json.encode({"items": gcp_sql_instances}))

    respx.get(
        f"https://sqladmin.googleapis.com/v1/projects/{gcp_project}/instances/{gcp_sql_instance_name}/users",
    ).respond(content=msgspec.json.encode({"items": gcp_sql_users}))

    respx.get(
        f"https://sqladmin.googleapis.com/v1/projects/{gcp_project}/instances/{gcp_sql_instance_name}/databases",
    ).respond(content=msgspec.json.encode({"items": gcp_sql_databases}))

    with patch_auth():
        resources = await explore_sql(gcp_project)

    assert len(resources) == 3



# mypy: disable-error-code="call-arg,index"
from typing import Any

import msgspec
from kubernetes import client

from lueur.links import add_link
from lueur.make_id import make_id
from lueur.models import Discovery, K8SMeta, Link, Resource
from lueur.platform.k8s.client import AsyncClient, Client
from lueur.rules import iter_resource

__all__ = ["explore_gateway"]


async def explore_gateway() -> list[Resource]:
    resources = []

    async with Client(client.CoreV1Api) as c:
        namespaces = await list_all_namespaces(c)

    async with Client(client.CustomObjectsApi) as c:
        for ns in namespaces:
            gateways = await explore_namespaced_gateways(c, ns, "v1")
            resources.extend(gateways)

            gateways = await explore_namespaced_gateways(c, ns, "v1beta1")
            resources.extend(gateways)

            gateways = await explore_namespaced_http_routes(c, ns, "v1")
            resources.extend(gateways)

            gateways = await explore_namespaced_http_routes(c, ns, "v1beta1")
            resources.extend(gateways)

    return resources


###############################################################################
# Private functions
###############################################################################
async def list_all_namespaces(c: AsyncClient) -> list[str]:
    response = await c.execute("list_namespace")

    namespaces = msgspec.json.decode(response.data)

    return [ns["metadata"]["name"] for ns in namespaces["items"]]


async def explore_namespaced_gateways(
    c: AsyncClient, ns: str, api_version: str = "v1"
) -> list[Resource]:
    response = await c.execute(
        "list_namespaced_custom_object",
        group="gateway.networking.k8s.io",
        version=api_version,
        plural="gateways",
        namespace=ns,
    )

    if response.status != 200:
        return []

    gateways = msgspec.json.decode(response.data)

    results = []
    for gw in gateways["items"]:
        meta = gw["metadata"]
        spec = gw["spec"]

        results.append(
            Resource(
                id=make_id(meta["uid"]),
                meta=K8SMeta(
                    name=f"{spec['gatewayClassName']}/{meta['name']}",
                    display=meta["name"],
                    kind="gateway",
                    platform="k8s",
                    ns=meta["namespace"],
                    category="loadbalancer",
                ),
                struct=gw,
            )
        )

    return results


async def explore_namespaced_http_routes(
    c: AsyncClient, ns: str, api_version: str = "v1"
) -> list[Resource]:
    response = await c.execute(
        "list_namespaced_custom_object",
        group="gateway.networking.k8s.io",
        version=api_version,
        plural="HTTPRoutes",
        namespace=ns,
    )

    if response.status != 200:
        return []

    routes = msgspec.json.decode(response.data)

    results = []
    for route in routes["items"]:
        meta = route["metadata"]

        results.append(
            Resource(
                id=make_id(meta["uid"]),
                meta=K8SMeta(
                    name=f"{meta['namespace']}/{meta['name']}",
                    display=meta["name"],
                    kind="httproute",
                    platform="k8s",
                    ns=meta["namespace"],
                    category="loadbalancer",
                ),
                struct=route,
            )
        )

    return results


def expand_links(d: Discovery, serialized: dict[str, Any]) -> None:
    for annotation in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='gateway'].struct.metadata.annotations['networking.gke.io/backend-services']",  # noqa E501
    ):
        r_id = annotation.parent.parent.parent.parent.obj["id"]  # type: ignore
        name = annotation.value.rsplit("/", 1)[-1]  # type: ignore

        p = f"$.resources[?@.meta.kind=='service' && @.struct.cloudRun.serviceName=='{name}']"  # noqa E501
        for svc in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="service",
                    path=svc.path,
                    pointer=str(svc.pointer()),
                ),
            )

            svc_name = svc.obj["meta"]["name"]  # type: ignore

            p = f"$.resources[?@.meta.kind=='slo' && match(@.meta.name, '{svc_name}/serviceLevelObjectives/.*')]"  # noqa E501
            for slo in iter_resource(serialized, p):
                add_link(
                    d,
                    r_id,
                    Link(
                        direction="out",
                        kind="slo",
                        path=slo.path,
                        pointer=str(slo.pointer()),
                    ),
                )

    for urlmap in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='gateway'].struct.metadata.annotations['networking.gke.io/url-maps']",  # noqa E501
    ):
        r_id = urlmap.parent.parent.parent.parent.obj["id"]  # type: ignore
        name = urlmap.value.rsplit("/", 1)[-1]  # type: ignore

        p = f"$.resources[?@.meta.kind=='global-urlmap' && @.meta.name=='{name}']"  # noqa E501
        for urlmap in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="global-urlmap",
                    path=urlmap.path,
                    pointer=str(urlmap.pointer()),
                ),
            )



from typing import Any

import msgspec
import pytest
import respx

from lueur.platform.gcp.monitoring import explore_monitoring

from tests.platform.gcp.utils import patch_auth


@respx.mock
@pytest.mark.asyncio
async def test_explore_monitoring(
    gcp_project: str,
    gcp_project_number: str,
    gcp_region: str,
    gcp_gke_cluster_name: str,
    gcp_monitoring_services: list[dict[str, Any]],
    gcp_monitoring_slo: list[dict[str, Any]],
    gcp_monitoring_alert_policies: list[dict[str, Any]],
    gcp_monitoring_notification_channels: list[dict[str, Any]],
) -> None:
    respx.post(
        "https://oauth2.googleapis.com/token",
    ).respond(content=msgspec.json.encode({"access_token": "xyz"}))

    respx.get(
        f"https://monitoring.googleapis.com/v3/projects/{gcp_project}/services",
    ).respond(
        content=msgspec.json.encode({"services": gcp_monitoring_services})
    )

    respx.get(
        f"https://monitoring.googleapis.com/v3/projects/{gcp_project_number}/services/ist:{gcp_project}-location-{gcp_region}-{gcp_gke_cluster_name}-default-demo/serviceLevelObjectives",
    ).respond(
        content=msgspec.json.encode(
            {"serviceLevelObjectives": [gcp_monitoring_slo[0]]}
        )
    )

    respx.get(
        f"https://monitoring.googleapis.com/v3/projects/{gcp_project_number}/services/demo-service/serviceLevelObjectives",
    ).respond(
        content=msgspec.json.encode(
            {"serviceLevelObjectives": [gcp_monitoring_slo[1]]}
        )
    )

    respx.get(
        f"https://monitoring.googleapis.com/v3/projects/{gcp_project}/alertPolicies",
    ).respond(
        content=msgspec.json.encode(
            {"alertPolicies": gcp_monitoring_alert_policies}
        )
    )

    respx.get(
        f"https://monitoring.googleapis.com/v3/projects/{gcp_project}/notificationChannels",
    ).respond(
        content=msgspec.json.encode(
            {"notificationChannels": gcp_monitoring_notification_channels}
        )
    )

    respx.get(
        f"https://monitoring.googleapis.com/v3/projects/{gcp_project}/groups",
    ).respond(content=msgspec.json.encode({"group": []}))

    with patch_auth():
        resources = await explore_monitoring(gcp_project)

    assert len(resources) == 6



import asyncio
import functools
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import Any

import httpx
from google.auth import exceptions, transport
from google.auth._credentials_async import Credentials
from google.auth._default_async import default_async
from google.oauth2._service_account_async import Credentials as OAuthCreds

__all__ = ["Client", "AuthorizedSession"]


class _Response(transport.Response):
    """
    Requests transport response adapter.
    Args:
        response (httpx.Response): The raw Requests response.
    """

    def __init__(self, response: httpx.Response):
        self._response = response
        self._content: bytes | None = None

    @property
    def status(self) -> int:
        return self._response.status_code

    @property
    def headers(self) -> httpx.Headers:
        return self._response.headers

    @property
    def data(self) -> bytes:
        return self._response.content

    async def raw_content(self):
        return await self.content()

    async def content(self) -> bytes:
        if self._content is None:
            self._content = await self._response.aread()
        return self._content


class Request(transport.Request):
    """Requests request adapter.
    This class is used internally for making requests using transports
    in a consistent way. If you use :class:`AuthorizedSession` you do not need
    to construct or use this class directly.
    This class can be useful if you want to manually refresh a
    :class:`~google.auth.credentials.Credentials` instance::
        import google.auth.transport.httpx_requests
        request = google.auth.transport.httpx_requests.Request()
        credentials.refresh(request)
    Args:
        client (httpx.Client): The client to use to make HTTP requests.
            If not specified, a session will be created.
    .. automethod:: __call__
    """

    def __init__(self, httpx_client: httpx.AsyncClient | None = None):
        self.client = httpx_client

    async def __call__(
        self,
        url: str,
        method: str = "GET",
        body: Any = None,
        headers: dict[str, str] | None = None,
        timeout: float = 180,
        **kwargs,
    ):
        """
        Make an HTTP request using httpx.
        Args:
            url (str): The URL to be requested.
            method (Optional[str]):
                The HTTP method to use for the request. Defaults to 'GET'.
            body (Optional[bytes]):
                The payload or body in HTTP request.
            headers (Optional[Mapping[str, str]]):
                Request headers.
            timeout (Optional[int]): The number of seconds to wait for a
                response from the server. If not specified or if None, the
                requests default timeout will be used.
            kwargs: Additional arguments passed through to the underlying
                requests :meth:`requests.Session.request` method.
        Returns:
            google.auth.transport.Response: The HTTP response.
        Raises:
            google.auth.exceptions.TransportError: If any exception occurred.
        """

        try:
            if self.client is None:  # pragma: NO COVER
                self.client = httpx.AsyncClient(http2=True)

            response = await self.client.request(
                method,
                url,
                data=body,
                headers=headers,
                timeout=timeout,
                **kwargs,
            )
            return _Response(response)

        except httpx.HTTPError as caught_exc:
            new_exc = exceptions.TransportError(caught_exc)
            raise new_exc from caught_exc


class AuthorizedSession(httpx.AsyncClient):
    """This is an implementation of the Authorized Session class. We utilize a
    httpx transport instance, and the interface mirrors the
    google.auth.transport.requests
    Authorized Session class, except for the change in the transport used in
    the use case.
    A Requests Session class with credentials.
    This class is used to perform requests to API endpoints that require
    authorization::
        from google.auth.transport import httpx_requests
        with httpx_requests.AuthorizedSession(credentials) as authed_session:
            response = await authed_session.request(
                'GET', 'https://www.googleapis.com/storage/v1/b')
    The underlying :meth:`request` implementation handles adding the
    credentials' headers to the request and refreshing credentials as needed.
    Args:
        credentials (google.auth._credentials.Credentials):
            The credentials to add to the request.
        refresh_status_codes (Sequence[int]): Which HTTP status codes indicate
            that credentials should be refreshed and the request should be
            retried.
        max_refresh_attempts (int): The maximum number of times to attempt to
            refresh the credentials and retry the request.
        refresh_timeout (Optional[int]): The timeout value in seconds for
            credential refresh HTTP requests.
        auth_request (google.auth.transport.httpx_requests.Request):
            (Optional) An instance of
            :class:`~google.auth.transport.httpx_requests.Request` used when
            refreshing credentials. If not passed,
            an instance of
            :class:`~google.auth.transport.httpx_requests.Request` is created.
        kwargs: Additional arguments passed through to the underlying
            ClientSession :meth:`httpx.ClientSession` object.
    """

    def __init__(
        self,
        credentials: Credentials | OAuthCreds,
        refresh_status_codes=transport.DEFAULT_REFRESH_STATUS_CODES,
        max_refresh_attempts=transport.DEFAULT_MAX_REFRESH_ATTEMPTS,
        refresh_timeout=None,
        auth_request=None,
        auto_decompress: bool = False,
        **kwargs,
    ):
        super(AuthorizedSession, self).__init__(**kwargs)
        self.credentials: Credentials = credentials
        self._refresh_status_codes = refresh_status_codes
        self._max_refresh_attempts = max_refresh_attempts
        self._refresh_timeout = refresh_timeout
        self._is_mtls = False
        self._auth_request = auth_request
        self._auth_request_session: httpx.AsyncClient | None = None
        self._auto_decompress = auto_decompress
        self._refresh_lock = asyncio.Lock()

    async def request(  # type: ignore
        self,
        method: str,
        url: str,
        data: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        max_allowed_time: float | None = None,
        timeout: float = 180.0,
        **kwargs,
    ) -> httpx.Response:
        """Implementation of Authorized Session httpx request.
        Args:
            method (str):
                The http request method used (e.g. GET, PUT, DELETE)
            url (str):
                The url at which the http request is sent.
            data (Optional[dict]): Dictionary, list of tuples, bytes, or
                file-like object to send in the body of the Request.
            headers (Optional[dict]): Dictionary of HTTP Headers to send with
                the Request.
            timeout (Optional[Union[float, httpx.ClientTimeout]]):
                The amount of time in seconds to wait for the server response
                with each individual request. Can also be passed as an
                ``httpx.ClientTimeout`` object.
            max_allowed_time (Optional[float]):
                If the method runs longer than this, a ``Timeout`` exception is
                automatically raised. Unlike the ``timeout`` parameter, this
                value applies to the total method execution time, even if
                multiple requests are made under the hood.
                Mind that it is not guaranteed that the timeout error is raised
                at ``max_allowed_time``. It might take longer, for example, if
                an underlying request takes a lot of time, but the request
                itself does not timeout, e.g. if a large file is being
                transmitted. The timout error will be raised after such
                request completes.
        """
        async with httpx.AsyncClient(http2=True) as self._auth_request_session:
            self._auth_request = Request(self._auth_request_session)

            _credential_refresh_attempt = kwargs.pop(
                "_credential_refresh_attempt", 0
            )
            request_headers = headers.copy() if headers is not None else {}

            auth_request = (
                self._auth_request
                if timeout is None
                else functools.partial(self._auth_request, timeout=timeout)
            )

            await self.credentials.before_request(
                auth_request, method, url, request_headers
            )

            response: httpx.Response = await super(
                AuthorizedSession, self
            ).request(
                method,
                url,
                data=data,
                headers=request_headers,
                timeout=timeout,
                **kwargs,
            )

            if (
                response.status_code in self._refresh_status_codes
                and _credential_refresh_attempt < self._max_refresh_attempts
            ):
                auth_request = (
                    self._auth_request
                    if timeout is None
                    else functools.partial(self._auth_request, timeout=timeout)
                )

                async with self._refresh_lock:
                    loop = asyncio.get_running_loop()
                    await loop.run_in_executor(
                        None, self.credentials.refresh, auth_request
                    )

                return await self.request(
                    method,
                    url,
                    data=data,
                    headers=headers,
                    timeout=timeout,
                    _credential_refresh_attempt=_credential_refresh_attempt + 1,
                    **kwargs,
                )

        return response


@asynccontextmanager
async def Client(
    base_url: str, creds: OAuthCreds | None = None
) -> AsyncIterator[AuthorizedSession]:
    credentials = creds
    if creds is None:
        credentials, _ = default_async()

    async with AuthorizedSession(credentials) as s:  # type: ignore
        s.base_url = httpx.URL(base_url)
        yield s



# mypy: disable-error-code="assignment,attr-defined,index,union-attr"
from concurrent.futures import ThreadPoolExecutor
from typing import Any

from lueur.links import add_link
from lueur.make_id import make_id
from lueur.models import AWSMeta, Discovery, Link, Resource
from lueur.platform.aws.client import Client
from lueur.rules import iter_resource

__all__ = ["explore_elbv2", "expand_links"]


def explore_elbv2(region: str) -> list[Resource]:
    resources = []

    lbs = explore_load_balancers(region)
    resources.extend(lbs)

    if not lbs:
        return resources

    futures = []
    with ThreadPoolExecutor(max_workers=len(lbs)) as executor:
        for lb in lbs:
            futures.append(
                executor.submit(explore_target_groups, region, lb.meta.name)
            )

            futures.append(
                executor.submit(explore_listeners, region, lb.meta.name)
            )

    for future in futures:
        resources.extend(future.result())

    return resources


###############################################################################
# Private functions
###############################################################################
def explore_load_balancers(region: str) -> list[Resource]:
    results = []

    with Client("elbv2", region) as c:
        lbs = c.describe_load_balancers()

        for lb in lbs["LoadBalancers"]:
            results.append(
                Resource(
                    id=make_id(lb["LoadBalancerArn"]),
                    meta=AWSMeta(
                        name=lb["LoadBalancerArn"],
                        display=lb["LoadBalancerName"],
                        kind="loadbalancer",
                        platform="aws",
                        region=region,
                        category="loadbalancer",
                    ),
                    struct=lb,
                )
            )

    return results


def explore_target_groups(region: str, lb_name: str) -> list[Resource]:
    with Client("elbv2", region) as c:
        tgs = c.describe_target_groups(
            LoadBalancerArn=lb_name,
        )

    results = []
    for tg in tgs["TargetGroups"]:
        results.append(
            Resource(
                id=make_id(tg["TargetGroupArn"]),
                meta=AWSMeta(
                    name=tg["TargetGroupArn"],
                    display=tg["TargetGroupName"],
                    kind="target-group",
                    platform="aws",
                    region=region,
                    category="loadbalancer",
                ),
                struct=tg,
            )
        )

    return results


def explore_listeners(region: str, lb_name: str) -> list[Resource]:
    with Client("elbv2", region) as c:
        listeners = c.describe_listeners(
            LoadBalancerArn=lb_name,
        )

    results = []
    for listener in listeners["Listeners"]:
        results.append(
            Resource(
                id=make_id(listener["ListenerArn"]),
                meta=AWSMeta(
                    name=listener["ListenerArn"],
                    display=listener["ListenerArn"],
                    kind="listener",
                    platform="aws",
                    region=region,
                    category="loadbalancer",
                ),
                struct=listener,
            )
        )

        results.extend(explore_rules(region, listener["ListenerArn"]))

    return results


def explore_rules(region: str, listener_arn: str) -> list[Resource]:
    with Client("elbv2", region) as c:
        rules = c.describe_rules(
            ListenerArn=listener_arn,
        )

    results = []
    for rule in rules["Rules"]:
        results.append(
            Resource(
                id=make_id(rule["RuleArn"]),
                meta=AWSMeta(
                    name=rule["RuleArn"],
                    display=rule["RuleArn"],
                    kind="listener-rule",
                    platform="aws",
                    region=region,
                    category="loadbalancer",
                ),
                struct=rule,
            )
        )

    return results


def expand_links(d: Discovery, serialized: dict[str, Any]) -> None:
    for lb_arn in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='loadbalancer'].meta.name",
    ):
        r_id = lb_arn.parent.parent.obj["id"]
        arn = lb_arn.value
        p = f"$.resources[?@.meta.kind=='listener' && @.struct.LoadBalancerArn=='{arn}']"  # noqa E501
        for listener in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="listener",
                    path=listener.path,
                    pointer=str(listener.pointer()),
                ),
            )

    for listener_arn in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='listener'].meta.name",
    ):
        r_id = listener_arn.parent.parent.obj["id"]
        arn = listener_arn.value
        prefix, suffix = arn.split("/", 1)
        rule_arn_prefix = f"{prefix}-rule/{suffix}"
        p = f"$.resources[?@.meta.kind=='listener-rule' && @.meta.name contains '{rule_arn_prefix}']"  # noqa E501
        for rule in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="listener-rule",
                    path=rule.path,
                    pointer=str(rule.pointer()),
                ),
            )

    for listener_rule_arn in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='listener-rule'].meta.name",
    ):
        r_id = listener_rule_arn.parent.parent.obj["id"]
        arn = listener_rule_arn.value
        prefix, suffix = arn.split("/", 1)
        listener_prefix = prefix.replace("listener-rule", "listener")
        listener_suffix, _ = suffix.rsplit("/", 1)
        listener_arn = f"{listener_prefix}/{listener_suffix}"
        p = f"$.resources[?@.meta.kind=='listener' && @.meta.name=='{listener_arn}']"  # noqa E501
        for listener in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="listener",
                    path=listener.path,
                    pointer=str(listener.pointer()),
                ),
            )

    for tg_arn in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='listener-rule'].struct.Actions.*.TargetGroupArn",  # noqa E501
    ):
        r_id = tg_arn.parent.parent.parent.parent.obj["id"]
        arn = tg_arn.value
        p = f"$.resources[?@.meta.kind=='target-group' && @.meta.name=='{arn}']"
        for tg in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="target-group",
                    path=tg.path,
                    pointer=str(tg.pointer()),
                ),
            )

    for subnet_id in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='loadbalancer'].struct.AvailabilityZones.*.SubnetId",
    ):
        r_id = subnet_id.parent.parent.parent.parent.obj["id"]
        subnet_id = subnet_id.value
        p = f"$.resources[?@.meta.kind=='subnet' && @.meta.name=='{subnet_id}']"
        for subnet in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="subnet",
                    path=subnet.path,
                    pointer=str(subnet.pointer()),
                ),
            )

    for lb_arn in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='target-group'].struct.LoadBalancerArns.*",
    ):
        tg = lb_arn.parent.parent.parent
        arn = lb_arn.value
        p = f"$.resources[?@.meta.kind=='loadbalancer' && @.meta.name=='{arn}']"
        for lb in iter_resource(serialized, p):
            r_id = lb.obj["id"]
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="target-group",
                    path=tg.path,
                    pointer=str(tg.pointer()),
                ),
            )



from pathlib import Path
from typing import Any, Iterable

import jsonpath

from lueur.models import Resource, Rule, Rules

__all__ = [
    "load_all_rules",
    "match_by_rule",
    "iter_resource",
    "get_at",
    "get_resource",
]


def load_all_rules(definitions: str | Path) -> Rules:
    return Rules.model_validate_json(Path(definitions).read_bytes())


def match_by_rule(resource: Resource, rule: Rule) -> Any | None:
    return jsonpath.findall(rule.matcher.path, resource.struct)


def iter_resource(
    discovery: dict[str, Any], path: str
) -> Iterable[jsonpath.JSONPathMatch]:
    yield from jsonpath.finditer(path, discovery)


def get_at(discovery: dict[str, Any], pointer: str) -> Any:
    return jsonpath.JSONPointer(pointer).resolve(discovery)


def get_resource(
    discovery: dict[str, Any], path: str
) -> jsonpath.JSONPathMatch | None:
    return jsonpath.match(path, discovery)
